/* Copyright (c) 2018 vesoft inc. All rights reserved.
 *
 * This source code is licensed under Apache 2.0 License,
 * attached with Common Clause Condition 1.0, found in the LICENSES directory.
 */

#include "base/Base.h"
#include "console/CmdProcessor.h"
#include "time/Duration.h"

namespace nebula {
namespace graph {

#define GET_VALUE_WIDTH(VT, FN, FMT) \
    VT val = col.get_ ## FN(); \
    size_t len = folly::stringPrintf(FMT, val).size(); \
    if (widths[idx] < len) { \
        widths[idx] = len; \
        genFmt = true; \
    }

void CmdProcessor::calColumnWidths(
        const cpp2::ExecutionResponse& resp,
        std::vector<size_t>& widths,
        std::vector<std::string>& formats) const {
    widths.clear();
    formats.clear();

    // Check column names first
    if (resp.get_column_names() != nullptr) {
        for (size_t i = 0; i < resp.get_column_names()->size(); i++) {
            widths.emplace_back(resp.get_column_names()->at(i).size());
        }
    }
    if (widths.size() == 0) {
        return;
    }

    if (resp.get_rows() == nullptr) {
        return;
    }

    std::vector<cpp2::ColumnValue::Type> types(
        widths.size(), cpp2::ColumnValue::Type::__EMPTY__);
    formats.resize(widths.size());
    // Then check data width
    for (auto& row : *(resp.get_rows())) {
        size_t idx = 0;
        for (auto& col : row.get_columns()) {
            CHECK(types[idx] == cpp2::ColumnValue::Type::__EMPTY__
                  || types[idx] == col.getType());
            bool genFmt = types[idx] == cpp2::ColumnValue::Type::__EMPTY__;
            if (types[idx] == cpp2::ColumnValue::Type::__EMPTY__) {
                types[idx] = col.getType();
            } else {
                CHECK_EQ(types[idx], col.getType());
            }

            switch (col.getType()) {
                case cpp2::ColumnValue::Type::__EMPTY__: {
                    break;
                }
                case cpp2::ColumnValue::Type::bool_val: {
                    // Enough to hold "false"
                    if (widths[idx] < 5UL) {
                        widths[idx] = 5UL;
                        genFmt = true;
                    }
                    if (genFmt) {
                        formats[idx] =
                            folly::stringPrintf(" %%%lds |", widths[idx]);
                    }
                    break;
                }
                case cpp2::ColumnValue::Type::integer: {
                    GET_VALUE_WIDTH(int64_t, integer, "%ld")
                    if (genFmt) {
                        formats[idx] =
                            folly::stringPrintf(" %%%ldld |", widths[idx]);
                    }
                    break;
                }
                case cpp2::ColumnValue::Type::id: {
                    // Enough to hold "0x{16 letters}"
                    if (widths[idx] < 18UL) {
                        widths[idx] = 18UL;
                        genFmt = true;
                    }
                    if (genFmt) {
                        formats[idx] =
                            folly::stringPrintf(" %%%ldLX |", widths[idx]);
                    }
                    break;
                }
                case cpp2::ColumnValue::Type::single_precision: {
                    GET_VALUE_WIDTH(float, single_precision, "%f")
                    if (genFmt) {
                        formats[idx] =
                            folly::stringPrintf(" %%%ldf |", widths[idx]);
                    }
                    break;
                }
                case cpp2::ColumnValue::Type::double_precision: {
                    GET_VALUE_WIDTH(double, double_precision, "%lf")
                    if (genFmt) {
                        formats[idx] =
                            folly::stringPrintf(" %%%ldlf |", widths[idx]);
                    }
                    break;
                }
                case cpp2::ColumnValue::Type::str: {
                    size_t len = col.get_str().size();
                    if (widths[idx] < len) {
                        widths[idx] = len;
                        genFmt = true;
                    }
                    if (genFmt) {
                        formats[idx] =
                            folly::stringPrintf(" %%%lds |", widths[idx]);
                    }
                    break;
                }
                case cpp2::ColumnValue::Type::timestamp: {
                    GET_VALUE_WIDTH(int64_t, timestamp, "%ld")
                    if (genFmt) {
                        formats[idx] =
                            folly::stringPrintf(" %%%ldld |", widths[idx]);
                    }
                    break;
                }
                case cpp2::ColumnValue::Type::year: {
                    if (widths[idx] < 4UL) {
                        widths[idx] = 4UL;
                        genFmt = true;
                    }
                    if (genFmt) {
                        formats[idx] = folly::stringPrintf(" %%%ldd |", widths[idx]);
                    }
                    break;
                }
                case cpp2::ColumnValue::Type::month: {
                    if (widths[idx] < 7UL) {
                        widths[idx] = 7UL;
                        genFmt = true;
                    }
                    if (genFmt) {
                        formats[idx] = folly::stringPrintf(" %%%ldd/%%02d |",
                                                           widths[idx] - 3);
                    }
                    break;
                }
                case cpp2::ColumnValue::Type::date: {
                    if (widths[idx] < 10UL) {
                        widths[idx] = 10UL;
                        genFmt = true;
                    }
                    if (genFmt) {
                        formats[idx] =
                            folly::stringPrintf(" %%%ldd/%%02d/%%02d |",
                                                widths[idx] - 6);
                    }
                    break;
                }
                case cpp2::ColumnValue::Type::datetime: {
                    if (widths[idx] < 26UL) {
                        widths[idx] = 26UL;
                        genFmt = true;
                    }
                    if (genFmt) {
                        formats[idx] =
                            folly::stringPrintf(" %%%ldd/%%02d/%%02d"
                                                " %%02d:%%02d:%%02d"
                                                ".%%03d%%03d |",
                                                widths[idx] - 22);
                    }
                    break;
                }
            }
            ++idx;
        }
    }
}
#undef GET_VALUE_WIDTH


void CmdProcessor::printResult(const cpp2::ExecutionResponse& resp) const {
    std::vector<size_t> widths;
    std::vector<std::string> formats;

    calColumnWidths(resp, widths, formats);

    if (widths.size() == 0) {
        return;
    }

    // Calculate the total width
    int32_t sum = 0;
    for (auto w : widths) {
        sum += w;
    }

    std::string headerLine(sum + 3 * widths.size() + 1, '=');
    std::string rowLine(sum + 3 * widths.size() + 1, '-');
    std::cout << headerLine << "\n|";

    printHeader(resp, widths);
    std::cout << headerLine << "\n";

    printData(resp, rowLine, widths, formats);
}


void CmdProcessor::printHeader(
        const cpp2::ExecutionResponse& resp,
        const std::vector<size_t>& widths) const {
    if (resp.get_column_names() == nullptr) {
        return;
    }

    size_t idx = 0;
    for (auto& cname : (*resp.get_column_names())) {
        std::string fmt = folly::stringPrintf(" %%%lds |", widths[idx++]);
        std::cout << folly::stringPrintf(fmt.c_str(), cname.c_str());
    }
    std::cout << "\n";
}


#define PRINT_FIELD_VALUE(...) \
    std::cout << folly::stringPrintf(formats[cIdx].c_str(), __VA_ARGS__)

void CmdProcessor::printData(const cpp2::ExecutionResponse& resp,
                             const std::string& rowLine,
                             const std::vector<size_t>& widths,
                             const std::vector<std::string>& formats) const {
    if (resp.get_rows() == nullptr) {
        return;
    }

    for (auto& row : (*resp.get_rows())) {
        int32_t cIdx = 0;
        std::cout << "|";
        for (auto& col : row.get_columns()) {
            switch (col.getType()) {
                case cpp2::ColumnValue::Type::__EMPTY__: {
                    std::string fmt = folly::stringPrintf(" %%%ldc |", widths[cIdx]);
                    std::cout << folly::stringPrintf(fmt.c_str(), ' ');
                    break;
                }
                case cpp2::ColumnValue::Type::bool_val: {
                    PRINT_FIELD_VALUE(col.get_bool_val() ? "true" : "false");
                    break;
                }
                case cpp2::ColumnValue::Type::integer: {
                    PRINT_FIELD_VALUE(col.get_integer());
                    break;
                }
                case cpp2::ColumnValue::Type::id: {
                    PRINT_FIELD_VALUE(col.get_id());
                    break;
                }
                case cpp2::ColumnValue::Type::single_precision: {
                    PRINT_FIELD_VALUE(col.get_single_precision());
                    break;
                }
                case cpp2::ColumnValue::Type::double_precision: {
                    PRINT_FIELD_VALUE(col.get_double_precision());
                    break;
                }
                case cpp2::ColumnValue::Type::str: {
                    PRINT_FIELD_VALUE(col.get_str().c_str());
                    break;
                }
                case cpp2::ColumnValue::Type::timestamp: {
                    PRINT_FIELD_VALUE(col.get_timestamp());
                    break;
                }
                case cpp2::ColumnValue::Type::year: {
                    PRINT_FIELD_VALUE(col.get_year());
                    break;
                }
                case cpp2::ColumnValue::Type::month: {
                    cpp2::YearMonth month = col.get_month();
                    PRINT_FIELD_VALUE(month.get_year(), month.get_month());
                    break;
                }
                case cpp2::ColumnValue::Type::date: {
                    cpp2::Date date = col.get_date();
                    PRINT_FIELD_VALUE(date.get_year(), date.get_month(), date.get_day());
                    break;
                }
                case cpp2::ColumnValue::Type::datetime: {
                    cpp2::DateTime dt = col.get_datetime();
                    PRINT_FIELD_VALUE(dt.get_year(),
                                      dt.get_month(),
                                      dt.get_day(),
                                      dt.get_hour(),
                                      dt.get_minute(),
                                      dt.get_second(),
                                      dt.get_millisec(),
                                      dt.get_microsec());
                    break;
                }
            }
            ++cIdx;
        }
        std::cout << "\n" << rowLine << "\n";
    }
}
#undef PRINT_FIELD_VALUE


bool CmdProcessor::processClientCmd(folly::StringPiece cmd,
                                    bool& readyToExit) {
    if (cmd == "exit") {
        readyToExit = true;
        return true;
    } else {
        readyToExit = false;
    }

    // TODO(sye) Check for all client commands

    return false;
}


void CmdProcessor::processServerCmd(folly::StringPiece cmd) {
    time::Duration dur;
    cpp2::ExecutionResponse resp;
    cpp2::ErrorCode res = client_->execute(cmd, resp);
    if (res == cpp2::ErrorCode::SUCCEEDED) {
        // Succeeded
        printResult(resp);
        if (resp.get_rows() != nullptr) {
            std::cout << "Got " << resp.get_rows()->size()
                      << " rows (Time spent: "
                      << resp.get_latency_in_us() << "/"
                      << dur.elapsedInUSec() << " us)\n";
        } else {
            std::cout << "Execution succeeded (Time spent: "
                      << resp.get_latency_in_us() << "/"
                      << dur.elapsedInUSec() << " us)\n";
        }
    } else if (res == cpp2::ErrorCode::E_SYNTAX_ERROR) {
        static const std::regex range("syntax error at 1.([0-9]+)-([0-9]+)");
        static const std::regex single("syntax error at 1.([0-9]+)");
        std::smatch result;
        auto *msg = resp.get_error_msg();
        auto verbose = *msg;
        if (std::regex_search(*msg, result, range)) {
            auto start = folly::to<size_t>(result[1].str());
            auto end = folly::to<size_t>(result[2].str());
            verbose = "syntax error near `" + std::string(&cmd[start-1], end - start + 1) + "'";
        } else if (std::regex_search(*msg, result, single)) {
            auto start = folly::to<size_t>(result[1].str());
            auto end = start + 8;
            end = end > cmd.size() ? cmd.size() : end;
            verbose = "syntax error near `" + std::string(&cmd[start-1], end - start + 1) + "'";
        }
        std::cout << "[ERROR (" << static_cast<int32_t>(res)
                  << ")]: " << verbose << "\n";
    } else {
        // TODO(sye) Need to print human-readable error strings
        auto msg = resp.get_error_msg();
        std::cout << "[ERROR (" << static_cast<int32_t>(res)
                  << ")]: " << (msg != nullptr ? *msg : "")
                  << "\n";
    }
}


bool CmdProcessor::process(folly::StringPiece cmd) {
    bool exit;
    if (processClientCmd(cmd, exit)) {
        return !exit;
    }

    processServerCmd(cmd);

    return true;
}

}  // namespace graph
}  // namespace nebula

